(**********************************************************
*
*  Datenbank fr Bume
*  
***********************************************************)

(*load "Set";
load "Map";
load "MyArray";
load "Polyhash";
load "Tree";*)

structure Database :> Database = struct
    
open Ref;
open Polyhash;
open Splayset;
open Splaymap;
open Tree;
open MyArray;

type ''a IList = ''a Tree.IList

exception Error of string
exception HashFind;

type id = int                (* als ID's werden integers verwendet *)

val hash_size = 262139       (* Primzahl in einem hoffentlich angemessenen Bereich *)
                             (* knapp unter 2^18 = 262144                          *)

(* Performance counters *)
(* Erlauben ein Profiling des Einfgens in die Datenbank *)

val c_tree2id = Ref 0
val c_handle_part = Ref 0
val c_expand_tree = Ref 0
val c_reduceLNodes = Ref 0
val c_representcycle = Ref 0
val c_representation = Ref 0
val c_scc = Ref 0

(* Die Signatur der Baeume in der Datenbank *)

datatype ''a dbsig = 
      L of ''a               (* Standardknoten *) 
    | D of ''a * id          (* Interner Datenbankknoten *)
    | P of id                (* New Strongly Connected Compoment *)

(* Die eigentliche Datenbank wird duch den folgenden Verbundtyp beschrieben *)
type ''a Database = 
    {tree2id : (''a dbsig DBLTree Ref, id(*, ''a Path*)) hash_table,    (* Baum -> ID *) 
     id2tree : (id, ''a dbsig DBLTree Ref) hash_table,                  (* ID -> Baum *)
     name2id : (string, id) hash_table,                                 (* Name -> ID *)
     id2name : (id, string) hash_table,                                 (* ID -> Name *)
     id_counter : int Ref,                                   (* Zaehler fuer die ID's *)
     l2int : ''a -> int,
     cmp : ''a dbsig * ''a dbsig -> order} 
                       
(* Liefert das Label eines Datenbankknotens *)
fun dblabel' (L l) = l
  | dblabel' (D(l,_)) = l
  | dblabel' _ = raise Error "dblabel"
fun dblabel t = dblabel'(getLabel(t))

(*********************************************************

  createDatabase: Erstellt eine neue Datenbank

**********************************************************)

fun string2int s = List.foldl (fn (x,a) => (Char.ord x + a) mod hash_size) 0 (explode s)

fun stringcompare s = case String.compare s of EQUAL => true
                      | _ => false

fun dbsig2int f x = 
    case x of (P i) => 45+i
    | _ => f (dblabel' x)

fun tree2id' c id f t =
    (c_tree2id := !c_tree2id + 1;
    c:= !c + 45; 
    case !t of
	E' => id := (!id + !c) mod hash_size
    |(N'{label=x,child=ch,done=done,...}) =>
	    if !done then ()
	    else (done := true;
		  id := (!id + !c + dbsig2int f x) mod hash_size;
		  map_array (tree2id' c id f) ch))

fun tree2id f t = getCyclesCount t (* Vorberechneten Wert holen *)
                (*let val id = Ref 0
                      val c = Ref 0
		  in  resetDBLTree t;
		      tree2id' c id f t;
		      resetDBLTree t;
		      !id
		  end*)

fun cmpLabel (P i, P j) = i = j
  | cmpLabel (P _, _) = false
  | cmpLabel (_, P _) = false
  | cmpLabel (x,y) = dblabel' x = dblabel' y
handle Error _ => raise Error "cmpLabel"

fun cmpdbsig cmp (a,b) = 
	    case (a,b) of
	        (P id, P id') => Int.compare(id,id')
	       |(P _, _) => LESS
	       |(_,P _) => GREATER
               |(L l, L l') => cmp(l,l')
	       |(L l, D(l',_)) => cmp(l,l')
	       |(D(l,_),L l') => cmp(l,l')
	       |(D(l,_),D(l',_)) => cmp(l,l') 

fun createDatabase cmp f :''a Database = 
	{ tree2id = Polyhash.mkTable (tree2id f, compareDBLTrees cmpLabel) (hash_size,HashFind),
          id2tree = mkTable (fn x => x, op=) (hash_size, HashFind),
          name2id = mkTable (string2int,stringcompare) (hash_size,HashFind),
          id2name = mkTable (fn x => x, op=) (hash_size,HashFind),
          id_counter = Ref 0,
          l2int = f, cmp = cmpdbsig cmp}

(* Liefert die Anzahl der ID's *)
fun dbgetids (db : ''a Database)= !(#id_counter db)

(********************************************************* 

  dbsetname: Weist dem Baum id einen Namen zu.
   
**********************************************************)
fun dbsetname id name (db : ''a Database) = 
     (Polyhash.insert (#id2name db) (id,name);
      Polyhash.insert (#name2id db) (name,id))

(********************************************************* 

  dbgetname: Liefert den Namen eines Baumes zurck
             oder NONE falls kein Name zugewiesen wurde. 

**********************************************************)
fun dbgetname id (db:''a Database) = 
   SOME(Polyhash.find (#id2name db) id)
handle HashFind => NONE 

(********************************************************* 

  dbsetname: Gibt die ID des mit name benannten Baumes 
             zurck.
   
**********************************************************)
fun dbgetidfromname name (db:''a Database) = 
    SOME (Polyhash.find (#name2id db) name) 
handle HashFind => NONE

(* Weist dem Baum t die ID id zu *)
fun dbinsert_id2tree id t (db:''a Database) = 
  Polyhash.insert (#id2tree db) (id,t) 

(* Liest den Baum aus der Datenbank. P-Knoten bleiben bestehen. *) 
fun dbgetdbtreefromid id (db:''a Database) = 
  SOME(Polyhash.find (#id2tree db) id)
handle HashFind => NONE 

(*********************************************************

  dbgettreefromid: Liest den zu ID gehoerigen Baum
                   vollstaendig aus der Datenbank aus

**********************************************************)
fun dbgettree' db done h t = 
    case Splaymap.peek(!done,t) of
	NONE =>
	    let val t' = case h of NONE => Ref E
		         |(SOME x) => x
	    in done := Splaymap.insert(!done,t,t');
	       case !t of
		   E' => t'
		 |(N'{label=D(l,id),child=ch,...}) =>
		       (t' := N(L l,List.map (dbgettree' db done NONE) (toList ch));t')
		 |(N'{label=P id,...}) =>
		       dbgettree' db (Ref (mkDict Ref.cmp)) (SOME t') (valOf(dbgetdbtreefromid id db))
		 | _ => raise Error "dbgettree'"
	    end
    |(SOME t') => t'
handle Option => (print "error: Option in dbgettree'"; raise Option)

fun dbgettreefromid id (db:''a Database) =
    let 
	val t = dbgetdbtreefromid id db
    in 
	case t of NONE => NONE
	    |(SOME t') => SOME(dbgettree' db (Ref (mkDict Ref.cmp)) NONE t')
    end

(*********************************************************

  dbgetid: Liefert die ID eines Baumes zurck bzw.
           NONE falls der Baum nicht in der Datenbank ist.

**********************************************************)
fun dbgetid (t:''a dbsig DBLTree Ref) (db:''a Database) = 
   case !t of 
       (N'{label=D(_,id),...}) => SOME id
     | (N'{label=P id,...}) => SOME id
     | _ => Polyhash.peek (#tree2id db) t 
handle HashFind => NONE 

(**************************************************************
*
* representation: Hauptalgorithmus, der den Baum in die
*  Datenbank eintrgt
*
**************************************************************)

(* Hasht einen Baum in die Datenbank *)
fun dbinsert' (db:''a Database) t =
    case !t of
	(N'{label=P r,...}) => r
      | _ =>
	    let 
		val id = !(#id_counter db)
	    in
		case Polyhash.peekInsert (# tree2id db) (t,id) of
		    NONE =>
			(#id_counter db := id + 1;
			 Polyhash.insert (# id2tree db) (id,t); 
			 id)
		  |(SOME oldid) => oldid
	    end

(* Hilfsfunktionen *)
(*fun map' t f a = 
    let val max = MyArray.size a
	val i = Ref 0
	val ret = Ref NONE
    in while !i < max do
	let val sh = f (sub(a,!i))
	in if sh <> NONE then (i:=max; ret:=sh)
	   else i := !i + 1
	end;
       !ret
    end*)

fun map'' f a0 a1  = 
    let val _ = if MyArray.size a0 <> MyArray.size a1 then raise Error "map''" else ()
	val max = MyArray.size a0
	val i = Ref 0
    in while !i < max do
	(f (sub(a0,!i)) (sub(a1,!i)); i := !i + 1)
    end

(*fun boolmap a0 a1 f = 
    if MyArray.size a0 <> MyArray.size a1 then false
    else let val max = MyArray.size a0
	     val i = Ref 0
	     val ret = Ref true
	 in while !i < max do
	     if f (sub(a0,!i)) (sub(a1,!i)) then i := !i + 1
	     else (ret := false; i:=max);
	    !ret 
	 end*)

(**************************************************************
   
   Vergleicht zwei Knoten t1 und t2 auf Aequivalenz. Dieser
   Algorithmus funktioniert NUR wenn t1 ein Knoten in einer
   Partition ueber t2 liegt und die Partition von t2 
   vollstaendig abgearbeitet ist. 

   Laufzeit: O(N1*N2*log(N1*N2))

*************************************************************) 
(*fun compareWithDone' done db t1 t2 = 
    (c_comparewithdone' := !c_comparewithdone' + 1;
     if member(!done,(t1,t2)) orelse
	member(!done,(t2,t1)) then true
     else case (!t1) of  
	 (N'{label=L l,child=xr,...}) =>
	     (case !t2 of
	      (N'{label=D(l',_),child=yr,...}) =>
		  (l = l') andalso 
		  (done := add(!done,(t1,t2)); 
		   boolmap xr yr (compareWithDone' done db))
	    |(N'{label=P id,...}) => false
	    | _ => raise Error "compareWithDone LL")
       | (N'{label=D(_,id),...}) =>
		  (case !t2 of
		       (N'{label=D(_,id'),...}) => id = id'
		     |(N'{label=P id',...}) => id = id'
		     | _ => raise Error "compareWithDone DL")
       | (N'{label=P id,...}) =>
		       (case !t2 of
			    (N'{label=D(_,id'),...}) => id = id'
			  |(N'{label=P id',...}) => id = id'
			  | _ => raise Error "compareWithDone DL")
       | E' => raise Error "compareWithDone E")

fun compareWithDone t1 t2 db = compareWithDone' (Ref (Splayset.empty Ref.cmpPair)) db t1 t2

*)
(**************************************************************************

   tryShare: Testet jeden Knoten in der unteren Partition
             auf quivalenz mit t

   Laufzeit: O(N_2) * O(compare_done)
             (N_2 = size(t'))
***************************************************************************)
(*
fun tryShare' db t t' = 
    (c_tryshare' := !c_tryshare' + 1;
     case !t' of
	E' => if compareWithDone t t' db then SOME t' else NONE
    |(N'{label=D(l,_),child=ch,done=done,...}) =>
	    if !done then NONE
	    else (done := true; 
		  if compareWithDone t t' db then SOME t'
		  else map' t (tryShare' db t) ch)
    |(N'{label=P _,...}) => NONE
    | _ => raise Error "tryShare")

fun tryShare t t' db = (resetDBLTree t'; tryShare' db t t')
*)

(* Merkt sich alle ID's der Partition sowie alle P Knoten *)

fun handlePartition idmap pmap restore t =

   mapDBLTreeSafe (fn x => (c_handle_part := !c_handle_part + 1;
         case !x of
	       (N'{label=D(_,id),...}) => 
		   (Polyhash.insert idmap (id,x);
		    case Polyhash.peek pmap id of
			NONE => ()
		      |(SOME lst) => (List.map (fn t => t := !x) lst;
				      restore := !restore @ lst;
				      Polyhash.insert pmap (id,nil)))
		   
	     | (N'{label=P id,...}) => 
		   (case Polyhash.peek idmap id of
			NONE => (case Polyhash.peek pmap id of
				     NONE => Polyhash.insert pmap (id,[x])
				   | (SOME lst) => Polyhash.insert pmap (id,x::lst))
		      |(SOME t') => (restore := x::(!restore); x := !t'))
	     | _ => raise Error "handlePartition")) t
     
handle Option => (print "handlepartition Option"; raise Option)


(* Sucht rekursiv die angrenzenden P Koten und hngt den entsprechenden Baum ein *)

fun expand_tree idmap pmap restore db t = 
    (c_expand_tree:= !c_expand_tree + 1;
     case !t of
	E' => t
    |(N'{label=L l,child=ch,done=done,...}) =>
	    if !done then t
	    else (done := true;
		  map_array_write (expand_tree idmap pmap restore db) ch;
                  t)
    |(N'{label=P id,done=done,...}) => 
		(case Polyhash.peek idmap id of
		     NONE => let val t' = valOf(dbgetdbtreefromid id db)
			     in handlePartition idmap pmap restore t'; t' end
		   |(SOME t') => t')

    |(N'{label=D _,...}) => raise Error "shareWithDone")
handle Option => (print "expand_tree Option"; raise Option)

(* Baut mit Hilfe der Blocks aus dem alten Baum einen neuen, reduzierten auf. *)
(* Dazu wird aus jedem Block ein Knoten als Reprsentant gewhlt und alle     *)
(* quivalenten Knoten durch diesen ersetzt.                                  *)
(* Es werden nur die L Knoten bearbeitet                                      *)

fun reduceLNodes t =
let
    val pmap = mkTable (fn x => x, op=) (20, HashFind)

    fun reduceLNodes' t =
	(c_reduceLNodes := !c_reduceLNodes + 1;
	 case !t of
	     E' => t
	   |(N'{label=P _,...}) => t
	   |(N'{label=D(l,id),...}) => 
		 (case Polyhash.peek pmap id of
		      NONE => let val t = newDBLNode(P id) 
			      in Polyhash.insert pmap (id,t); t end 
			|(SOME t) => t)
	   |(N'{label=L l,...}) => 
		 (* In welchem Block steht der Knoten? *)
		 case getBlock t of
		     NONE     => raise Error "uninitialized node"
		   | (SOME b) => 
			 (* Haben wir schon einen Knoten aus diesem Block benutzt? *)
			 (case block_get_leader b of
			 NONE      => (block_set_leader(b,SOME t);
				       (case !t of 
					    (N'{child=child,...}) => map_array_write reduceLNodes' child
					  | _         => ());
					    t)
		       | (SOME t') => t')) 
in
    reduceLNodes' t
end

fun find_D_node trees =
let 
    val node = Ref(IList.first trees)
    val ret = Ref NONE
in while not(IList.isEnd(!node)) andalso !ret = NONE do
    let val t = IList.value(!node)
	val _ = c_handle_part := !c_handle_part + 1
    in  
	case !t of 
	    (N'{label=D _,...}) => ret := SOME t 
	  | _ => ();
		node := IList.next(!node)
    end;
    !ret
end
handle Option => (print "find_D_node Option"; raise Option)

(****************************************************************************

  shareWithDone: Sucht einen aequivalenten Knoten in einer angrenzenden
                 Partition. Dazu werden die angrenzenden Partitionen
		 angehngt und dann mit Share alles minimiert.

  Laufzeit: O(N^2*log(N))

*****************************************************************************) 

fun shareWithDone db t' = 
    let
	val restore = Ref nil
	(* Hier werden die P Knoten notiert *)
	val pmap =  mkTable (fn x => x, op=) (30, HashFind)

	(* Zu jeder id auf die man stt wird der entsprechende Knoten gemerkt *)
	val idmap = mkTable (fn x => x, op=) (30, HashFind)

	val t = copyDBLTree t'
	(*val _ = resetDBLTree t*)

	(* Die Partitionen anhngen *)
	val _ = expand_tree idmap pmap restore db t

	(* Die Rckverweise in dem DBLTree mssen neu berechnet werden *)
        val _ = computeBackPointers t

	(* Den Partitioning Algorithmus laufen lassen, aber
	 noch nicht Sharen!!! *)
	val _ = shareDBLTreeWithDone (#cmp db) t

	(* Whrend des Algorithmus wurden die P Knoten der Datenbank
	 berschrieben. In restore sind alle Daten fr eine Rekonstrution enthalten, so
	 dass die Datenbank hier wieder rekonstruiert werden kann. *)
        val _ = List.map (fn x => case !x of (N'{label=D(_,id),...}) => x := !(newDBLNode(P id))
		             | _ => raise Error "shareWithDone 3") (!restore)
	
    in
	(* Den Block des Knotens holen *)
	 case getBlock t of NONE => raise Error "shareWithDone 2"
	    |(SOME blk) => case blk of 
		(B blk) => case !blk of 
		    {trees=trees,...} => 
			(* Schauen ob in dem Block ein D Knoten enthalten ist ... *)
			case find_D_node trees of
			    NONE => (reduceLNodes t; (t,false))   (* falls nein, liegt kein 
								   root unfolding vor *)
			  |(SOME t') => (t',true)                 (* falls ja liegt root unfolding vor *)
    end

(* Ersetzt in einer Partition alle Knoten durch P Knoten. Dazu
   wird ein quivalenter Knoten in einer schon abgearbeiteten 
   Partition bentigt *)

fun writeIDtoNodes db t swdt = 
    case (!t,!swdt) of
      (N'{label=D _,...},_) => ()
    | (N'{label=P _,...},_) => ()
    | (N'{child=ch,...},N'{child=ch',...}) => 
	    (t := !(newDBLNode(P(valOf(dbgetid swdt db))));
	    map'' (writeIDtoNodes db) ch ch')
    | _ => raise Error "writeIDtoNodes"
handle Option => (print "error: Option in writeIDtoNodes\n"; raise Option)


(*****************************************************************************
*
* computeHashNumbers: Berechnet die Hash Zahlen vor 
*
******************************************************************************)

fun computeHashNumbers f t =
  let 
    val n = Ref 0
    val h = Ref 0
    val nodes = Ref nil

    fun reset f t = 
	mapDBLTree (fn t => 
           case !t of E' => () | (N'{label=l,cycles_count=cycles,t1=t1,...}) => 
                (t1 := 0; cycles := (f l))) t

    fun addChildren t = 
      let
         val c = Ref 0
         val i = Ref 1
      in
         case !t of E' => () | (N'{child=ch,t1=t1,...}) =>
            map_array (fn x => case !x of E' => () | (N'{cycles_count=cycles,...}) => 
                 (h := (!h + !t1) mod hash_size;
                  t1 := (!t1 + 13*(!i+73)*(!cycles)) mod hash_size; 
                  i := !i + 1)) ch         
      end

    fun updateCycles t = case !t of 
	    E' => () 
         | (N'{cycles_count=cycles,t1=t1,...}) => 
                (c_tree2id := !c_tree2id + 1;
                 cycles := (!cycles + !t1 + !h) mod hash_size;
	         t1 := 0)
  in
    reset f t;
    mapDBLTree (fn t' => (n := !n + 1;
                         nodes := t'::(!nodes))) t;

    while !n > 0 do
      (List.map addChildren (!nodes);
       List.map updateCycles (!nodes);
       h := 0;
       n := !n div 2)
  end
handle Option => (print "error: Option in computeHashNumbers\n"; raise Option)

(*****************************************************************************

   representCycle: Bearbeitet eine gefundene Partition.

   Laufzeit: O(swd) + O(share) + O(N*(N+K)*log(N))

******************************************************************************)

fun representCycle db t =

    let
        val _ = c_representcycle:= !c_representcycle + 1; 

       	(*val _ = resetDBLTree t*)
	val (t',is_swd) = shareWithDone db t  (* schauen ob der Knoten schon in einer tieferen
                                                 Partition existiert *) 
    in
	if is_swd then writeIDtoNodes db t t'     (* falls ja dann alle Knoten mit den bereits 
                                                      gefundenen Partitionen ersetzen *)
        else ( computeHashNumbers (dbsig2int(#l2int db)) t';
	       
	        case dbgetid t' db of

                  (* Wenn die Partition schon in der Datenbank ist dann alle Knoten durch die
                     Partitionen in der Datenbank ersetzen. *)

		    (SOME id) => writeIDtoNodes db t (valOf(dbgetdbtreefromid id db))

                  (* Wenn die Partition noch nicht in der Datenbank ist dann muessen alle Knoten in 
                     die Datenbank eingefuegt werden. *)
                
	          |  NONE     => (mapDBLTree (fn x => case !x of (N'{label=P _,...}) => ()
	                                      | _ => setLabel (D(dblabel x,dbinsert' db x)) x) t';
	                         writeIDtoNodes db t t')
            )
   end
handle Option => (print "error: Option in representCycle\n"; raise Option)

(*********************************************************

  scc_find

  !!!!!! benutze cycles_count zum Speichern des dfs_count
  !!!!!! done sagt, ob der Knoten schon einer scc zugeteilt wurde

  Laufzeit: O(N+K) + O(rep_cycle)
            N = Knoten
            K = Kanten

**********************************************************)

fun scc f t =
    let
       	val roots = Ref nil 
	val dfs_count = Ref 0    (* zaehlt die Knoten durch, an denen man vorbeikommt *)
	val unfinished = Ref nil (* hier war ich schon, habe aber noch nichts getan   *)

	fun insert_comp_numbers v = 
	    let val w = hd(!unfinished)
	    in  unfinished := tl(!unfinished);
		case !w of 
		    (*(N'{cycles_count=cc,...}) => cc := !comp_count*)
		    (N'{done=done,...})       => done := true
		  |  E'                       => raise Error "scc_dfs";
	        if Ref.cmp(w,v) <> EQUAL then insert_comp_numbers v
		else ()
	    end

	fun isPartition t =
	    isDone t
	    orelse case !t of
		(N'{label=P _,...}) => true
	      | _                   => false

	(* loescht alle vermuteten Wurzeln, die erst nach w getroffen wurden,
	   da sie in derselben Partition liegen und damit doch keine Wurzeln sind *)
	fun delete_roots_below w =
	    if not (null(!roots)) andalso getCyclesCount (hd(!roots)) > getCyclesCount w then
		(roots := tl(!roots);
		 delete_roots_below w)
	    else ()
	    
	fun explore v w = 
	    if getCyclesCount w = ~1 then scc_dfs w
	    else if not (isPartition w) then delete_roots_below w else ()
	    
	and scc_dfs v =
	    case !v of
		(N'{done=done,child=child,cycles_count=cc,...}) =>
		    (c_scc := !c_scc + 1;
		     cc := !dfs_count;
		     dfs_count := !dfs_count + 1;
		     
		     (*done := true;*)
		     unfinished := v::(!unfinished);
		     roots := v::(!roots);
		     map_array (explore v) child;
		     if v=hd(!roots) then
			(insert_comp_numbers v;
			 roots := tl(!roots);
			 f v)
		     else () )
	      | E'=> raise Error "scc"
    in
       resetDBLTree t;        (* Den Baum zunaechst initialisieren *)
       scc_dfs t
    end

(***************************************************************************

   representation: 

 ****************************************************************************)  

fun representation db t =
  case !t of (N'{label=P r,...}) => r
  | _ => 
    (
     c_tree2id := 0;
     c_handle_part := 0;
     c_expand_tree:= 0;
     c_reduceLNodes := 0;
     c_representcycle := 0;
     c_representation := 0;
     c_scc := 0;

     scc (representCycle db) t;             (* Komponenten suchen und in Datenbank eintragen *)
     case !t of             
	 (N'{label=P(r),...}) => r            (* ID zurueckgeben *)
 	| _ => raise Error "Internal Database Error (no Partition found)"
    )
before ((*print ("c_tree2id         = " ^ Int.toString (!c_tree2id) ^ "\n");
        print ("c_handle_part     = " ^ Int.toString (!c_handle_part) ^ "\n");
	print ("c_expand_tree     = " ^ Int.toString (!c_expand_tree) ^ "\n");
	print ("c_reduceLNodes    = " ^ Int.toString (!c_reduceLNodes) ^ "\n");
        print ("c_representcycle  = " ^ Int.toString (!c_representcycle) ^ "\n");
	print ("c_representation  = " ^ Int.toString (!c_representation) ^ "\n");
        print ("c_scc             = " ^ Int.toString (!c_scc) ^ "\n\n")*))
handle Option => raise Error "Internal Database Error (Option)"


(*********************************************************

  dbinsert: Fgt einen Baum zur Datenbank hinzu 
            und liefert seine ID zurck.

**********************************************************)

fun dbinsert (db:''a Database) t = 
    representation db (copyDBLTree t)

end















