(**********************************************************************
*
*              Kontextfreie Syntax fr Rekursive Typen
*
***********************************************************************)

structure Parser :> Parser = struct

open Database;
open Lex;
open Tree;

exception Error of string;

(* Abstrakte Syntax fr rekursive Typen *)
datatype ty = 
  Null 
| Unit 
| Arrow of ty*ty
| App of ty*ty
| Add of ty*ty
| Mul of ty list
| Id of string
| My of string*ty
| Lambda of string*ty

(*****************************************************************
 
  ty: Wandelt eine token list in einen Typ um

  Kontextfreie Grammatik fr Typen:

  ty      = [arrowty] ty
  arrowty = addty [-> ty]
  addty   = [addty +] multy
  multy   = atty [* multy]
  atty    = null | unit | x | My t. M | Lambda t. M | ( ty )

*****************************************************************)

fun atty nil = raise Error "atty: wrong type 1"
  | atty (NULL::ts) = (ts,Null)
  | atty (UNIT::ts) = (ts,Unit)
  | atty ((ID s)::ts) = (ts, Id s)
  | atty (MY::(ID s)::DOT::tr) = let val (tr',b) =  ty' tr 
                                 in (tr',My(s,b)) end
  | atty (LAMBDA::(ID s)::DOT::tr) = let val (tr',b) =  ty' tr 
				     in (tr',Lambda(s,b)) end
  | atty (LPAR::ts) = (case ty' ts of
                        (RPAR::tr,b) => (tr,b)
                        | (_,_) => raise Error "atty: wrong type 2")
  | atty _ = raise Error "atty: wrong type 3"

and multy nil =  raise Error  "multy: Error wrong type" 
  | multy t = multy' t nil

and multy' nil l = (nil,Mul(l))
  | multy' t l = (case atty t of
                  (MUL::tr,b) => multy' tr (l@[b])
                 |(tr,b) => if null l then (tr,b)
                            else (tr,Mul(l@[b])))

and addty nil = raise Error "addty"
 |  addty t = let val (tr,b) = multy t
              in addty' tr b end

and addty' (ADD::tr) last = let val (tr',b) = multy tr
                                val t' = Add(last,b)
                             in addty' tr' t' end
  | addty' tr last = (tr,last)

and arrowty nil = raise Error "arrowty"
 |  arrowty tr = case addty tr of
                  (ARROW::tr,b) => let val (tr',b') = arrowty tr 
				   in (tr',Arrow(b,b')) end
		 |(tr,b) => (tr,b)

and ty' nil = raise Error "ty: wrong type"
  | ty' ts = case arrowty ts of (tr,b) => ty'' tr b

and ty'' nil last = (nil,last)
  | ty'' (RPAR::tr) last = (RPAR::tr,last)
  | ty'' tr last = let val (tr',b) = arrowty tr
		   in ty'' tr' (App(last,b)) end

fun ty ts = case ty' ts of
      (nil,t) => SOME t
     |(_,_) => NONE
handle Error s => (print (s ^ "\n"); NONE)


(*****************************************************************

  ty2str: Wandelt einen Typ in einen String um.

*****************************************************************)

fun ty2str_atty Null = "null"
  | ty2str_atty Unit = "unit"
  | ty2str_atty (Id s) = s
  | ty2str_atty (My(id,t)) = "(" ^ "u " ^ id ^ "." ^ ty2str_ty t ^ ")"
  | ty2str_atty (Lambda(id,t)) = "(" ^ "l " ^ id ^ "." ^ ty2str_ty t ^ ")"
  | ty2str_atty t = "(" ^ (ty2str_ty t) ^ ")"

and ty2str_multy (Mul tr) = ty2str_multy' tr
  | ty2str_multy t = ty2str_atty t

and ty2str_multy' nil = ""
  | ty2str_multy' (t::nil) = ty2str_atty t
  | ty2str_multy' (t::tr) = (ty2str_atty t) ^ "*" ^ (ty2str_multy' tr)

and ty2str_addty (Add(t1,t2)) = (ty2str_addty t1) ^ "+" ^ (ty2str_multy t2)
  | ty2str_addty t = ty2str_multy t

and ty2str_arrowty (Arrow(t1,t2)) = (ty2str_addty t1) ^ "->" ^ (ty2str_arrowty t2)
  | ty2str_arrowty t = ty2str_addty t

and ty2str_ty (App(t1,t2)) = (ty2str_arrowty t1) ^ " " ^ (ty2str_ty t2)
  | ty2str_ty t = ty2str_arrowty t

fun ty2str t = ty2str_ty t


end

