(* punktierte Graphen mit Labels *)
(* Da                                                                           *)
(* 1. jeder Baum durch den Knoten an der Wurzel reprsentiert wird              *)
(* 2. Jeder Knoten gleichzeitig den bei ihm beginnenden Unterbaum reprsentiert *)
(* werdem im Folgenden Bume und Knoten synonym verwendet.                      *)

signature Tree = sig

  type 'a ListNode
  type 'a IList
  type ''a array
  eqtype 'a Ref
  datatype ''a Tree = E | N of ''a * ''a Tree Ref list

  (* Datentyp eines doppelt verknpften Baumes *)
  datatype ''a DBLTree = 
      E' | 
      N' of {label: ''a,                                    (* Label des Knotens                          *)
	     block: ''a block option Ref,                   (* Block, der den Knoten enthlt              *)
	     blocknode: ''a DBLTree Ref ListNode option Ref,(* ListNode des Knotens in diesem Block       *)
	     child: ''a DBLTree Ref array,                  (* Unterbume                                 *)
	     pred:  ''a DBLTree Ref list array,             (* Vorgnger                                  *)
	     incoming: int array,                           (* Anzahl der Vorgnger                       *)
	     done: bool Ref,                                (* Knoten in der Prozedur schon abgearbeitet? *)
	     cycles_count: int Ref, t1: int Ref,            (* Wird in Repr bentigt                      *)
	     private_done: int Ref }			    (* Internes Done                              *)

  (* Ein Block fasst mehrere Knoten zu einer Einheit zusammen.     *)
  (* Nach einer Verfeinerung enthlt jeder Block genau alle        *)
  (* zu einem Knoten quivalenten Knoten.                          *)
  (* Jeder Block speichert die Knoten, die Anzahl deren Unterbume *)
  (* sowie einen Reprsentanten, der beim Zusammenbauen verwendet  *)
  (* wird.                                                         *)
  (* Die Komponente "split" wird bentigt, um beim Verfeinern      *)
  (* zwischen im Block verbleibenden und zu verschiebenden Knoten  *)
  (* zu unterscheiden.                                             *)    
  and ''a block = B of {trees : ''a DBLTree Ref IList,         (* Knoten in dem Block                     *)
			split : ''a DBLTree Ref ListNode IList,(* Knoten, die nach dem Teilen in einen    *)
			                                       (* neuen Block verschoben werden           *)
			leader : ''a DBLTree Ref option Ref,   (* ausgezeichneter Knoten zum "Schrumpfen" *)
			inToTest: bool array,                  (* steckt der Block in ToTest(i)?          *)
			incoming: int array } Ref              (* Zahl der eingehenden Kanten mit Label i *)

  (* Zugriff auf das Label eines Knoten *)
  val getLabel : ''a DBLTree Ref -> ''a
  val setLabel : ''a -> ''a DBLTree Ref -> unit

  (* Zugriff auf den Reprsentanten eines Blocks *)
  val block_get_leader : ''a block -> ''a DBLTree Ref option
  val block_set_leader : ''a block * ''a DBLTree Ref option -> unit

  (* Zugriff auf den Block eines Knotens *)
  val getBlock : ''a DBLTree Ref -> ''a block option
  val setBlock : ''a DBLTree Ref -> ''a block option -> unit

  (* Zugriff auf den Zustand "abgearbeitet" eines Knoten *)
  val isDone : ''a DBLTree Ref -> bool
  val setDone : bool -> ''a DBLTree Ref -> unit

  (* Berechnet die Rckverweise neu *)
  val computeBackPointers : ''a DBLTree Ref -> unit

  (* Zugriff auf die Anzahl der zu einem Knoten schlieenden Zykel *)
  val getCyclesCount: ''a DBLTree Ref -> int
  val setCyclesCount: int -> ''a DBLTree Ref -> unit

  (* Konvertierung in einen doppelt verknpften Baum und zurck *)
  val toDBLTree : ''a Tree Ref -> ''a DBLTree Ref
  val fromDBLTree : ''a DBLTree Ref -> ''a Tree Ref

  (* Erstellung eines Knotens *)
  val newDBLNode : ''a -> ''a DBLTree Ref
  (* Setzt bei einem doppelt verknpften Baum Done und CyclesCount zurck *)
  val resetDBLTree : ''a DBLTree Ref -> unit

  (* Map auf DBLTrees *)
  val mapDBLTree : (''a DBLTree Ref -> 'b) -> ''a DBLTree Ref -> unit
  val mapDBLTreeSafe : (''a DBLTree Ref -> 'b) -> ''a DBLTree Ref -> unit

  (* Minimiert einen DBL Baum *)
  val shareDBLTree : (''a*''a->order) -> ''a DBLTree Ref -> ''a DBLTree Ref
  val shareDBLTreeWithDone : (''a*''a->order) -> ''a DBLTree Ref -> unit

  (* Testet Bume auf semantische Gleichheit *)
  val eqTree : ''a Tree Ref -> ''a Tree Ref -> bool

  (* Minimiert den Baum *)
  val share : (''a*''a->order) -> ''a Tree Ref -> ''a Tree Ref

  (* Spezifikation des Share Algorithmus   *)
  (* Stellt eine naive Implementierung dar *)
  val share_spec : ''a Tree Ref -> ''a Tree Ref

  (* "Zeichnet" den Baum *)
  val drawTree : ''a Tree Ref -> (''a->string) -> string

  (* Kopierfunktion *)
  val copyTree : ''a Tree Ref -> ''a Tree Ref
  val copyDBLTree : ''a DBLTree Ref -> ''a DBLTree Ref

  (* Liefert das Label an der Wurzel *)
  val label: ''a Tree Ref -> ''a

  (* Erstellt einen Zufallsgraph                           *)
  (* Angegeben werden                                      *)
  (*   Eine Funktion zum erstellen von Labels aus Integers *)
  (*   Die maximale Anzahl Knoten im Graph                 *)
  (*   Die maximale Anzahl Nachfolger je Knoten            *)
  (*   Eine obere Schranke fr die Beschriftungen          *)
  val randomTree : (int -> ''a) -> (''a -> bool) -> int -> int -> int -> ''a Tree Ref

  (* Vergleicht zwei Bume syntaktisch *)
  val compareTrees : ''a Tree Ref * ''a Tree Ref -> bool
  val compareDBLTrees : (''a*''a->bool) -> ''a DBLTree Ref * ''a DBLTree Ref -> bool

  (* Map-Funktion auf Bumen *)
  val nodeMapTree: (''a Tree Ref -> 'b) -> ''a Tree Ref -> unit
  val mapTree: (''a Tree Ref -> ''b) -> ''a Tree Ref -> ''b Tree Ref

  (* Liefert fr einen Baum                           *)
  (*   Die Anzahl der Knoten                          *)
  (*   Die Anzahl der Kanten                          *)
  (*   Die maximale Zahl an Nachfolgern eines Knotens *)
  val getTreeSize : ''a Tree Ref -> int * int * int

end where type ''a array = ''a MyArray.array
      and type 'a Ref = 'a Ref.Ref
      and type 'a IList = 'a IList.IList
      and type 'a ListNode = 'a IList.ListNode







