(***************************************************************
*
* Darstellung eines zyklischen Baumes 
* 
****************************************************************)

(*load "Set";
load "Map";
load "IList";
load "Hash";
load "Ref";
load "Random";
load "Int";*)

structure Tree :> Tree = struct

open Ref;
open IList;
open Splayset;
open Splaymap;
open Random;
open MyArray;

(* Neuer Zufallszahlengenerator *)
val gen = newgen();
exception Error of string;

(* Datentyp eines einfach verknpften Baumes *)
datatype ''a Tree = E | N of ''a * ''a Tree Ref list

(* Datentyp eines doppelt verknpften des Baumes *)
datatype ''a DBLTree = 
    E' | 
    N' of {label: ''a,                                     (* Label des Knotens                          *)
	   block: ''a block option Ref,                    (* Block, der den Knoten enthlt              *)
	   blocknode: ''a DBLTree Ref ListNode option Ref, (* ListNode des Knotens in diesem Block       *)
	   child: ''a DBLTree Ref array,                   (* Unterbume                                 *)
	   pred:  ''a DBLTree Ref list array,              (* Vorgnger                                  *)
	   incoming: int array,                            (* Anzahl der Vorgnger                       *)
	   done: bool Ref,                                 (* Knoten in der Prozedur schon abgearbeitet? *)
	   cycles_count: int Ref, t1: int Ref,             (* Wird in Repr bentigt                      *)
	   private_done: int Ref }			   (* Internes Done                              *)

(* jeder Block speichert die Knoten, die Anzahl deren Unterbume
   sowie den beim Zusammenbauen verwendeten Knoten *)    
and ''a block = B of { trees : ''a DBLTree Ref IList,         (* Knoten in dem Block                     *)
		       split : ''a DBLTree Ref ListNode IList,(* Knoten, die nach dem Teilen in einen    *)
			                                      (* neuen Block verschoben werden           *)
		       leader : ''a DBLTree Ref option Ref,   (* ausgezeichneter Knoten zum "Schrumpfen" *)
		       inToTest: bool array,                  (* steckt der Block in ToTest(i)?          *)
		       incoming: int array } Ref              (* Zahl der eingehenden Kanten mit Label i *)

(*********************************************************************************************************)
(* Zugriffsfunktionen auf Bumen                                                                         *)	
(*********************************************************************************************************)

fun label t = case !t of
                  (N(l,_)) => l
		|  E       => raise Error "label" 

fun numberOfSubtrees t = case !t of
    (N(_,st)) => length(st)
  | E         => 0

fun newDBLNode l = 
  Ref(N' {label=l,
	  block=Ref NONE,
	  blocknode=Ref NONE,
	  pred=array(0,nil),
          incoming=array(0,0),
	  child=array(0,Ref E'),
	  done=Ref false,
	  cycles_count = Ref 0, 
	  t1 = Ref 0,
	  private_done = Ref 0});

fun getLabel t =
    case !t of
	E' => raise Error "getLabel"
      |(N'{label=l,...}) => l

fun setLabel l t =
  case !t of
      E' => ()
  |(N'{block=blk, 
	  blocknode=bnode, 
	  child=ch, 
	  pred=pred, 
	  incoming=inc, 
	  done=done, private_done = private_done,
	  cycles_count=cycles,t1=t1,...}) => t:= N'{label=l,
					  block=blk, 
					  blocknode=bnode, 
					  child=ch, 
					  pred=pred, 
					  incoming=inc, 
					  done=done, private_done = private_done, 
					  cycles_count=cycles,t1=t1}

fun getChild t =
    case !t of
	E' => raise Error "getChild"
      |(N'{child=ch,...}) => ch

fun getPred t =
    case !t of
	E' => raise Error "getPred"
      |(N'{pred=p,...}) => p

fun resetPred t =
    case !t of
	E' => ()
    |(N'{pred=p,...}) => map_array_write (fn _ => nil) p

fun getIncoming t = 
    case !t of 
	E' => array(0,0)
      |(N'{incoming=incoming,...})=> incoming

fun resetIncoming t =
    case !t of
	E' => ()
    |(N'{incoming=inc,...}) => map_array_write (fn _ => 0) inc

fun setCyclesCount b t = case !t of
    E' => ()
  |(N'{cycles_count=c,...}) => c := b

fun getCyclesCount t = case !t of
    E' => ~1
  |(N'{cycles_count=c,...}) => !c

fun setDone b t = case !t of
    E' => ()
  |(N'{done=done,...}) => done := b
	
fun isDone t = case !t of
    E' => true
  |(N'{done=done,...}) => !done

fun getBlock n = case !n of E' => raise Error "getBlock"
| (N'({block = blk, ...})) => !blk
    
fun setBlock n blk = case !n of E' => ()
| (N'({block = b, ...})) => b := blk

fun getBlockNode n = case !n of E' => raise Error "getBlockNode"
| (N'({blocknode = n, ...})) => !n
    
fun setBlockNode n node = case !n of E' => ()
| (N'({blocknode = blocknode, ...})) => blocknode := node
        
fun numberOfSubtrees t = 
  case !t of
    E' => 0
  |(N'{child=child,...}) => MyArray.size child
   

(*********************************************************************************************************)
(* Zugriffsfunktionen auf Blcken                                                                        *)
(*********************************************************************************************************)

fun block_num_incoming i (B blk) = 
    case !blk of 
	{incoming=incoming,...} => sub(incoming,i)
	    
fun block_trees (B blk) = 
    case !blk of 
	{trees=trees,...} => trees
	    
fun block_in_wait i (B blk) = 
    case !blk of 
	{inToTest=wait,...} => sub(wait,i)
	   
	    
fun block_get_leader (B blk) = 
    case !blk of 
	{leader=leader,...} => !leader
	    
fun block_set_leader((B blk),l) = 
    case !blk of 
	{leader=leader,...} => leader := l
	    
fun block_remove_wait i (B blk) =
    case !blk of
	{inToTest=wait,...} => update(wait,i,false)

fun block_get_trees (B blk) =
	    case !blk of {trees=trees,...} => trees

fun block_size_inc (B blk) = 
	case !blk of {incoming=incoming,...} => 
	    MyArray.size incoming 

fun block_get_split (B blk) =
	    case !blk of {split=split,...} => split

fun block_del_split blk n = erase(block_get_split blk,n)

fun block_empty_split blk = IList.clear (block_get_split blk)

fun block_can_split (B blk) =
	    case !blk of {split=split,...} => (IList.size split) <> 0

(*********************************************************************************************************)
(* Funktionen auf Blcken                                                                                *)
(*********************************************************************************************************)

fun block_create () = 
    B(Ref {trees = IList.empty(),
	   split = IList.empty(),
	   leader = Ref NONE,
	   inToTest = array(0,true),
	   incoming = array(0,0)}) 

(* Lschen eines Knotens aus dem Block *)
fun block_del (B blk) node = 
    case !blk of 
	{trees=trees, incoming=incoming,...} => (setBlock (value node) NONE;
						 setBlockNode (value node) NONE;
						 erase(trees,node);
						 op_array(op-,incoming,getIncoming(value(node))))
	    
(* Einfgen eines Knotens in den Block *)
fun block_insert (B blk) t = 
    case !blk of
	{trees=trees, incoming=incoming,...} => (setBlock t (SOME(B blk));
						 setBlockNode t (SOME(pushBack(trees,t)));
						 op_array(op+,incoming,getIncoming(t)))

(* Markiert einen Knoten als abzuspalten *)
fun block_move_to_split bl n = 
    if isDone (value n) then ()
    else (setDone true (value n);
	  pushBack(block_get_split bl, n);())

(* Teilt einen Block auf indem die Knoten in "split" *)
(* In einen neuen Block verschoben werden.           *)
fun block_do_split (bl as B blk) = 
     let
	 val new = block_create()
     in
	 case !blk of
	     {trees=trees,split=split,...} => 
		 if IList.size trees = IList.size split 
		     then (imap (setDone false o value) trees;
			   block_empty_split bl;
			   bl)
		 else if IList.isEmpty split then bl
		 else (imap (fn n => (block_del bl (value n);
				      block_insert new (value(value(n)));
				      setDone false (value(value(n))))) split;
		       block_empty_split bl;
		       new)
     end


(**************************************************************
*
* nodeMapTree: map-Funktion auf Bumen,
*              die auf dem gegebenen Baum arbeitet
*
**************************************************************)

fun nodeMapTree' done f t =
    if member(!done, t) then ()
    else 
	(done := add(!done,t);
	 case !t of
	     (N(l,xs)) => (f t; List.map (nodeMapTree' done f) xs; ())
	   |  E        => ()
	)

fun nodeMapTree f t = nodeMapTree' (Ref (Splayset.empty Ref.cmp)) f t

(**************************************************************
*
* mapTree: map-Funktion auf Bumen,
*          die den Baum unverndert lsst und einen neuen erstellt
*
**************************************************************)

fun mapTree' done f t =
    case peek(!done, t) of (SOME t') => t'
    | NONE => let val t' = Ref E
              in done := insert(!done,t,t');
	      case !t of
	        (N(l,xs)) => t' := N(f t, List.map (mapTree' done f) xs)
	      |  E        => ();
              t'
              end

fun mapTree f t = mapTree' (Ref (mkDict Ref.cmp)) f t

(*************************************************************
*
* mapDBLTree: Mapping fr DBL Bume 
*
*************************************************************)
fun mapDBLTree f t =
    let
	val True = case !t of (N'{private_done=pdone,...}) => !pdone + 1
                             | _ => 0
	fun mapDBLTree' f t =
	    case !t of
		(N'{child=child,private_done=pdone,...}) => 
		    if !pdone = True then () 
		    else (pdone := True; f t; 
			  map_array (mapDBLTree' f) child; ())
	      |  E'        => ()
    in  mapDBLTree' f t
    end

(*************************************************************
*
* mapDBLTreeSafe: Mapping fr DBL Bume 
*
*************************************************************)
fun mapDBLTreeSafe f t =
    let
	val done = Ref(Splayset.empty Ref.cmp)

	fun mapDBLTree' f t =
	    case !t of
		(N'{child=child,...}) => 
		    if member(!done,t) then () 
		    else (done := add(!done,t);
			  f t; 
			  map_array (mapDBLTree' f) child; ())
	      |  E'        => ()
    in  mapDBLTree' f t
    end

(* setzt Done und CyclesCount zurck *)
fun resetDBLTree t = mapDBLTree (fn x => (setDone false x; 
                                          setCyclesCount ~1 x)) t;


(************************************************************
*
* toDBLTree: Wandelt einen normalen Baum 
*            in einen DBL Baum um
*
************************************************************)

(* edge speichert, ber welche Kante der Knoten   *)
(* erreicht wurde, parent ist der Ausgangsknoten. *)
(* Gab es keinen Vorgnger (z.B. bei der Wurzel), *)
(* so ist edges = ~1 gesetzt.                     *)

fun toDBLTree' done (edge,parent) t = 
    case peek(!done,t) of 
	NONE =>  (* Der Knoten wurde noch nicht umgewandelt. *)
	    (case !t of 
		 E        => Ref E'
	       |(N(x,tr)) => 
		     let
			 val t'= newDBLNode x
			 val _ = done := insert(!done,t,t');
			 fun f i t = (i:= !i+1; toDBLTree' done (!i-1,t') t)
			 val pr = getPred t'
			 val inc = getIncoming t'
			 val ch = getChild t'
		     in
			 if edge <> ~1 then (update(pr,edge,[parent]);
					     update(inc,edge,1))
			               else ();
			 copy_array(fromList(List.map (f (Ref 0)) tr,Ref E'),ch);
			 t'
		     end
		 )
      |(SOME t') =>  (* Der Knoten wurde schon in t' umgewandelt *)
		 ( case !t' of
		       E' => t'
		     |(N'{pred=prd,incoming=inc,...}) => (update(prd,edge,parent::sub(prd,edge));
							  update(inc,edge,1+sub(inc,edge));
							  t')
		 )

fun toDBLTree t = toDBLTree' (Ref (mkDict Ref.cmp)) (~1,Ref E') t

(************************************************************
*
* fromDBLTree: Wandelt einen DBL Baum 
*              in einen normalen Baum um
*
************************************************************)

(* Hier werden einfach berschssige Angaben gelscht. *)

fun fromDBLTree' done t = 
    case peek(!done,t) of
	(SOME t') => t'
      | NONE => let 
		    val t' = Ref E
		in done:=insert(!done,t,t');
		   case !t of
		       E' => t'
		     |(N'{label=l,child=child,...}) => (t':= N(l,List.map (fromDBLTree' done) (toList child));t')
		end

fun fromDBLTree t = fromDBLTree' (Ref (mkDict Ref.cmp)) t

(************************************************************
*
* computeBackPointers: Berechnet die Rckverweise eines 
*                      DBL Baumes neu 
*
************************************************************)

fun computeBackPointers' done (edge,parent) t = 
    case !t of
	E' => ()
      |(N'{pred=prd,incoming=inc,child=ch,...}) => 
	    (if edge <> ~1 then (update(prd,edge,parent::sub(prd,edge));
				 update(inc,edge,1+sub(inc,edge))) 
	                   else (); 
	     if Splayset.member(!done,t) then ()
	     else let
		      fun f i t' = (i:= !i+1; computeBackPointers' done (!i-1,t) t')
		  in
		      done := add(!done, t);
		      map_array  (f (Ref 0)) ch
		  end) 

fun computeBackPointers t = 
(
 mapDBLTreeSafe (fn t => (resetPred t; resetIncoming t)) t;
 computeBackPointers' (Ref (Splayset.empty Ref.cmp)) (~1,Ref E') t
)

(**************************************************************
*
* eqTree: Berechnet, ob zwei Bume semantisch gleich sind
*
* Laufzeit: O(N^2*log(N))
*
**************************************************************)

fun eqTree'' nil nil _ = true
  | eqTree'' nil  _  _ = false
  | eqTree'' _  nil  _ = false
  | eqTree'' (x::xr) (y::yr) done = eqTree' x y done andalso eqTree'' xr yr done

and eqTree' t1 t2 done = 
 if member(!done,(t1,t2)) orelse
    member(!done,(t2,t1)) then true
 else case (!t1,!t2) of
       (E,E) => true
     | (N(l,xr),E) => false
     | (E,N(l',yr)) => false
     | (N(l,xr),N(l',yr)) => (l = l') andalso 
                             (done := add(!done,(t1,t2)); eqTree'' xr yr done)

fun eqTree t1 t2 = eqTree' t1 t2 (Ref (Splayset.empty Ref.cmpPair))

(**************************************************************
*
* share: Minimiert einen Baum
*
* Laufzeit: O(K*k*log(N)) = O(k^2*N*log(N))
*           K = Anzahl Kanten
*           k = maximale Anzahl Nachfolger eines Knotens
*           N = Anzahl der Knoten
*
**************************************************************)

   
(* Stellt einen Block in die Warteliste *)
fun block_to_wait waiting i (B blk) = 
    case !blk of
	{inToTest=wait,incoming=incoming,...} => 
	    if sub(incoming,i) = 0 then ()
	    else (update(wait,i,true);
		  pushBack(waiting,(i,B blk)); ())
		
(* Initialisierung der Waiting Queue *)
fun block_init_wait waiting blk = 
    let val i = Ref 0
	val max = block_size_inc blk
    in while !i < max do
	(block_to_wait waiting (!i) blk;
	 i := !i + 1)
    end			      

fun initWaiting blocks waiting = imap ((block_init_wait waiting) o value) blocks
    
fun initiateBlocks blocks (f:(''a*''a)->order) n  =
    let
	(* Findet den ersten Block, in dem ein Knoten mit dem geg. 
	 Label und #Nachfolger eingetragen ist*)
	fun cmp((x,y),(x',y')) = 
	    case Int.compare(y,y') of 
		EQUAL => f(x,x')
	      | c => c
		    
	val label2block: (''a * int, ''a block) dict Ref = Ref (Splaymap.mkDict cmp)
	    
	(* Sortiert die Knoten des Baumes nach Labels und Zahl der 
	 Unterbume geordnet in 'blocks' ein *)
	fun initiateBlocks' n =
	    if isDone n then () 
	    else let 
		     val p = (getLabel n,numberOfSubtrees n)
		     val labelInBlock = Splaymap.peek(!label2block,p)
		 in  
		     setDone true n;	(* Knoten ist bearbeitet *)
		     (case labelInBlock of
			  NONE => let val newblock = block_create()  (* Neuen Block erstellen *)
				  in  
				      block_insert newblock n;       (* Knoten zu Block hinzufgen *)
				      pushBack(blocks, newblock);    (* Block zu den Blcken werfen *)
				      label2block := Splaymap.insert(!label2block,p,newblock)
				  end
			| (SOME b) => block_insert b n);
		     map_array initiateBlocks' (getChild n)
		 end
    in
	initiateBlocks' n
    end
 
(* Verfeinert die Partitionierung, bis nur noch quivalente Knoten  *)
(* in demselben Block gespeichert sind                              *)
fun handleBlock waiting blk i = 
    let
	(* Diese Blcke wurden beim Zurcklaufen angetroffen *)
	val toDo = IList.empty()
	    
	(* Fgt einen Knoten zur Splitliste des Blocks hinzu *)
	fun move_to_new_block t =  
	    case !t of 
		E' => ()
	      |(N'{block=bl, blocknode=node,...}) => 
		    (case !bl of NONE => ()
		  |(SOME b) => (if block_can_split b then () else (pushBack(toDo,b);());
				    block_move_to_split b (valOf(!node))))
			 
	(* Verfolgt die Kante i des Knotens n zurck *)
	fun follow_back n = 
	    case !(value(n)) of
		E' => ()
	      |(N'{pred=prd,...}) => 
		    (List.map (fn t => (case getBlock t of NONE => ()
		  | (SOME b) => move_to_new_block t)) 
		     (sub(prd,i));())
		    
	(* Splittet den Block blk *)
	fun do_split blk = 	
	    let
		val j = Ref 0
		val max = block_size_inc blk 
		val newblk = block_do_split blk
	    in
		if newblk = blk then ()               (* Der Block wurde nicht gesplittet. *)
		else while !j < max do
		    (if block_in_wait (!j) blk then   (* Wenn der Block in der Warteliste war,         *)
			 block_to_wait waiting (!j) newblk   (* dann mssen nun beide Teile in die Warteliste.*)
		     else if block_num_incoming (!j) blk < block_num_incoming (!j) newblk 
			      then block_to_wait waiting (!j) blk     (* Anderenfalls muss nur der kleinere Teil   *)
			          else block_to_wait waiting (!j) newblk; (* in die Warteliste.                        *)
				      j := !j + 1)
	    end
	
    in
	block_remove_wait i blk;                    (* Nach der Kante wurde verfeinert *)
	imap follow_back (block_trees blk);         (* Kanten zurcklaufen             *)
	imap (do_split o value) toDo                (* und nun die Blcke teilen       *)
    end

fun handle_wait waiting n = (case value(n) of (i,blk) => handleBlock waiting blk i;
		             erase(waiting,n);())

fun splitBlocks waiting = imap (handle_wait waiting) waiting

(* Baut mit Hilfe der Blocks aus dem alten Baum einen neuen, reduzierten auf. *)
(* Dazu wird aus jedem Block ein Knoten als Reprsentant gewhlt und alle     *)
(* quivalenten Knoten durch diesen ersetzt.                                  *)
fun reduceTree (t:''a DBLTree Ref) =
    (* In welchem Block steht der Knoten? *)
    case getBlock t of
	NONE     => raise Error "uninitialized node"
      | (SOME b) => 
	    (* Haben wir schon einen Knoten aus diesem Block benutzt? *)
	    (case block_get_leader b of
		 NONE      => (block_set_leader(b,SOME t);
			       (case !t of 
				    (N'{child=child,...}) => map_array_write reduceTree child
				  | _         => ());
				    t)
	       | (SOME t') => t') 
		 
fun shareDBLTree cmp t: (''a DBLTree Ref) = 
let 
    (* Blcke, nach denen noch verfeinert werden muss *)
    val waiting:(int*''a block) IList = IList.empty()
	
    (* Partitionierung der Knoten eines Baumes *)
    val blocks: (''a block) IList = IList.empty()
in
 (* Baum initialisieren *)	
 resetDBLTree t;
 
 (* Initiale Blcke berechnen *)
 initiateBlocks blocks cmp t;
 
 (* Waiting Queue initialisieren *)
 initWaiting blocks waiting;
 
 (* Blcke verfeinern *)
 resetDBLTree t;	
 splitBlocks waiting;
 
 (* Baum minimieren *)
 reduceTree t
end

fun share cmp t = fromDBLTree(shareDBLTree cmp (toDBLTree t))

fun shareDBLTreeWithDone cmp t = 
let
    (* Blcke, nach denen noch verfeinert werden muss *)
    val waiting:(int*''a block) IList = IList.empty()
	
    (* Partitionierung der Knoten eines Baumes *)
    val blocks = IList.empty()
in
 (* Baum initialisieren *)	
 resetDBLTree t;
 
 (* Initiale Blcke berechnen *)
 initiateBlocks blocks cmp t;
 
 (* Waiting Queue initialisieren *)
 initWaiting blocks waiting;
 
 (* Blcke verfeinern *)
 resetDBLTree t;	
 splitBlocks waiting;
 ()     
end
 
(*******************************************************

 Spezifikation des Share Algorithmus

 Laufzeit: O((K+N)*N*N^2*log(N)) = O(k*N^4*log(N))
           K = Anzahl Kanten
           N = Anzahl Knoten
	   k = maximale Anzahl Nachfolger eines Knotens

*******************************************************)

fun find nil f = NONE
  | find (x::xr) f = if f x then SOME x else find xr f
  
fun share' done t = 
    case find (!done) (eqTree t) of 
	(SOME t') => t'
      | NONE      => (done:=t::(!done); 
		      case !t of 
			  E => ()
			| (N(l,xr)) => t := N(l,List.map (share' done) xr);
		      t)
fun share_spec t = share' (Ref nil) t


(**************************************************************
*
* drawTree: Zeichnet einen Baum
*
**************************************************************)

fun drawTree'' nil l2s map i = ""
  | drawTree'' (t::nil) l2s map i = drawTree' t l2s map i
  | drawTree'' (t::tr) l2s map i = (drawTree' t l2s map i) ^ "," ^ drawTree'' tr l2s map i

and drawTree' t l2s map i = case peek(!map,t) of
    NONE => (i := !i+1; map:=insert(!map,t,!i-1); 
             (case !t of
                E => "E"

              |(N(l,xr)) => (l2s l) ^ "(" ^ (drawTree'' xr l2s map i) ^ ")"))
  | (SOME id) => "node" ^ Int.toString(id) 

fun drawTree t l2s  = drawTree' t l2s (Ref (mkDict Ref.cmp)) (Ref 0)

(**************************************************************
*
* copyTree: Erstellt eine Kopie eines Baumes
*
**************************************************************)

fun copyTree' m t =
  case peek(!m,t) of
    NONE => let val t' = Ref E
            in m:=insert(!m,t,t'); 
               case !t of
                (N(l,tr)) => t':=N(l, List.map (copyTree' m) tr)
               | E => t':=E;
               t'
            end
  |(SOME t') => t'

fun copyTree t = copyTree' (Ref (mkDict Ref.cmp)) t

fun copyDBLTree t = (toDBLTree o fromDBLTree) t

(**************************************************************
*
* randomTree: Erstellt einen Zufallsbaum 
*
**************************************************************)

fun randomNodes f 0 l = nil
  | randomNodes f n l = (Ref (N(f(trunc(random gen * real(l))),nil)))::(randomNodes f (n-1) l)

fun nth n nil = raise Error "nth"
  | nth 0 (x::_) = x
  | nth n (x::xs) = nth (n-1) xs

fun randSubList' 0 l lst = nil
  | randSubList' n l lst = 
    let val s = nth (trunc(random gen * real(l))) lst
    in s :: randSubList' (n-1) l lst end
fun randSubList n lst = randSubList' n (length lst) lst 

fun linkNodes' ispart k (nodes:(''a Tree Ref list)) nil  = nil
  | linkNodes' ispart k nodes ((x::xs):(''a Tree Ref list)) = 
    let
	val subtrees = trunc (random gen * real(k+1))
    in
	(case !x of
	    E => ()
	  | (N(v,tr)) => if ispart v then x := (N(v,nil))
                         else x := (N(v,randSubList subtrees nodes)));
	linkNodes' ispart k nodes xs
    end
fun linkNodes ispart k nodes = (linkNodes' ispart k nodes nodes; nodes)


(* Erstellt einen zuflligen Baum mit maximal
    - n Knoten
    - k Kanten pro Knoten
    - Knotenbeschriftungen von 0 bis l-1
   f erstellt aus einer Zahl ein Label
*)
fun randomTree f ispart n k l = if n<=0 then hd(linkNodes ispart k (randomNodes f 1 l)) 
			        else hd(linkNodes ispart k (randomNodes f n l)) 

(**************************************************************
*
* compareTrees: Vergleicht zwei Bume (nur syntaktisch)
*
**************************************************************)

fun mymap nil nil f = true
  | mymap (x::xr) (y::yr) f = (f x y andalso mymap xr yr f)
  | mymap _ _ _ = false

fun compareTrees' cmp done0 done1 link t0 t1 = 
    let 
	val m0 = member(!done0,t0) 
	val m1 = member(!done1,t1)
    in
	if m0 andalso m1 andalso (member(!link,(t0,t1))) then true
	else 
	    if (not m0 andalso not m1) then
		(done0 := add(!done0,t0);
		 done1 := add(!done1,t1);
                 link := add(!link,(t0,t1));
		 case (!t0,!t1) of
		     (E,E) => true
		   |(N(l0,tr0),N(l1,tr1)) => cmp(l0,l1) andalso
			 mymap tr0 tr1 (compareTrees' cmp done0 done1 link)
		   | _ => false)
	    else false
    end

fun compareTrees (t0,t1) = compareTrees' op= 
	                  (Ref (Splayset.empty Ref.cmp)) 
                          (Ref (Splayset.empty Ref.cmp)) 
                          (Ref (Splayset.empty Ref.cmpPair)) t0 t1 

fun compareDBLTrees cmp (t0,t1) =
    compareTrees' cmp 
      (Ref (Splayset.empty Ref.cmp)) 
      (Ref (Splayset.empty Ref.cmp))
      (Ref (Splayset.empty Ref.cmpPair))
      (fromDBLTree t0) (fromDBLTree t1)


(* Berechnet die Gre eines Baumes *)
fun getTreeSize' done (nodes,edges,maxedge) t =
    if member(!done,t) then ()
    else (done := add(!done,t);
	  case !t of
	      E => nodes := !nodes + 1
	    |(N(_,tr)) => (nodes := !nodes + 1;
			   edges := !edges + length tr;
			   maxedge := Int.max(!maxedge,length tr);
			   List.map (getTreeSize' done (nodes,edges,maxedge)) tr; ()))

fun getTreeSize (t: ''a Tree Ref) = 
    let val (nodes,edges,maxedge) = (Ref 0, Ref 0, Ref 0)
    in getTreeSize' 
	(Ref (Splayset.empty Ref.cmp)) (nodes,edges,maxedge) t;
       (!nodes,!edges,!maxedge)
    end

end

