structure TypeTrees :> TypeTrees =
struct

exception Error of string;

open Ref;
open Parser;
open Tree;
open Splaymap;
open Splayset;
open Database;

datatype TyLabel = null_ | unit_ | arrow_ | 
                   add_ | mul_ | id_ of string | 
		   lambda_ | back_ of int

(* Wandelt ein Label in einen String um. *)
fun label2string' (unit_) = "unit"
  | label2string' (null_) = "null"
  | label2string' (arrow_) = "arrow"
  | label2string' (add_) = "add"
  | label2string' (mul_) = "mul"
  | label2string' (lambda_) = "l"
  | label2string' (back_ i) = "back"^Int.toString i
  | label2string' (id_ s) = "Id(" ^ s ^ ")"

fun label2string (L x) = label2string' x
  | label2string (P id) = Int.toString(id)
  | label2string (D(x,id)) = "D("^label2string' x^","^Int.toString id^")"

(* Wandelt ein Label in einen Int um.               *)
(* Diese Funktion wird spter zum Hashen verwendet. *)
fun label2int null_ = 1
  | label2int unit_ = 2
  | label2int arrow_ = 3 
  | label2int add_ = 4 
  | label2int mul_ = 5
  | label2int (id_ s) = 6 + List.foldr (fn (x,a) => Char.ord x + a) 0 (explode s)
  | label2int lambda_ = 7
  | label2int (back_ x) = 8 + x

fun labelcmp (id_ s1,id_ s2) = String.compare(s1,s2) 
  | labelcmp (id_ _, _) = LESS
  | labelcmp (_, id_ _) = GREATER
  | labelcmp (a,b) = Int.compare(label2int a,label2int b)


(*****************************************************************

  tytree: Wandelt einen Typ in einen zykischen Typenbaum um.
          Dabei werden Applikationen gleich mit erledigt.

*****************************************************************)

(* Schlieen von Zyklen, d.h.    *)
(* Umwandlung z.B. von x.y*x in *)
(*        * <-.                  *)
(*       / \__|                  *)
(*      y                        *)
fun closecycles' t0 s done t = if member(!done,t) then t else 
   (done:=add(!done,t); (case !t of
    (N(L(arrow_),tr))  => (t:=N(L(arrow_), List.map (closecycles' t0 s done) tr); t)
   |(N(L(add_),tr))    => (t:=N(L(add_),   List.map (closecycles' t0 s done) tr); t)
   |(N(L(mul_),tr))    => (t:=N(L(mul_),   List.map (closecycles' t0 s done) tr); t)
   |(N(L(lambda_),tr)) => (t:=N(L(lambda_),List.map (closecycles' t0 s done) tr); t)
   |(N(L(id_ s'),_)) => if s=s' then t0 else t
   | _ => t))
fun closecycles t t0 s = closecycles' t0 s (Ref (Splayset.empty Ref.cmp)) t

(* Binden von Variablen an das bindende Lambda *)
fun bindback' s l done t = if member(!done,t) then t else 
   (done:=add(!done,t); (case !t of
    (N(L(arrow_),tr)) => (t:=N(L(arrow_),List.map (bindback' s l done) tr); t)
   |(N(L(add_),tr)) => (t:=N(L(add_),List.map (bindback' s l done) tr); t)
   |(N(L(mul_),tr)) => (t:=N(L(mul_),List.map (bindback' s l done) tr); t)
   |(N(L(lambda_),tr)) => (t:=N(L(lambda_),List.map (bindback' s (l+1) done) tr); t)
   |(N(L(id_ s'),_)) => if s=s' then (t:=N(L(back_ l),nil);t) else t
   | _ => t))
fun bindback t s = bindback' s 0 (Ref (Splayset.empty Ref.cmp)) t

(* Wendet einen Typ auf einen anderen an. *)
(* Beispiel: (l x.x*nat) unit -> unit*nat *)
fun subst' t2 l done t = 
    if member(!done,t) then t 
    else 
	(done:=add(!done,t);
	 (case !t of
	      (N(L(arrow_),tr))  => (t:=N(L(arrow_), List.map (subst' t2 l done) tr); t)
	    | (N(L(add_),tr))    => (t:=N(L(add_),   List.map (subst' t2 l done) tr); t)
	    | (N(L(mul_),tr))    => (t:=N(L(mul_),   List.map (subst' t2 l done) tr); t)
	    | (N(L(lambda_),tr)) => (t:=N(L(lambda_),List.map (subst' t2 (l+1) done) tr); t)
	    | (N(L(back_ l'),_)) => if l'=l then t2 else t
	    | _ => t))
fun subst t2 t1 = 
    case !t1 of
	(N(L(lambda_),[child])) => subst' t2 0 (Ref (Splayset.empty Ref.cmp)) child
      | _                       => raise Error "subst: Applikation ohne Lambda"

(* Umwandlung Typ -> Typenbaum *)
(* Beispiel:                   *)
(*   l x.x*unit                *)
(*   -> Lambda x0              *)
(*          |                  *)
(*         Mul                 *)
(*        /   \                *)
(*       x0  unit              *)
fun tytree' db Null = Ref (N(L(null_),nil))
  | tytree' db Unit = Ref (N(L(unit_),nil))
  | tytree' db (Arrow(t1,t2)) = Ref (N(L(arrow_),[tytree' db t1,tytree' db t2]))
  | tytree' db (App(t1,t2)) = (* wende t2 auf t1 an, falls der mit einem Lambda beginnt*)
    let val t1' = tytree' db t1
	val t2' = tytree' db t2
    in
	case !t2' of
	    (N(P id,nil)) => subst t1' (valOf(dbgettreefromid id db))
  	  | (N(L(lambda_),[c])) => subst t1' t2'
	  | _ => raise Error "tytree': App without Lambda"
    end
  | tytree' db (Add(t1,t2)) = Ref (N(L(add_),[tytree' db t1, tytree' db t2]))
  | tytree' db (Mul lr) = Ref (N(L(mul_),List.map (tytree' db) lr))
  | tytree' db (My(s,t)) = 
       let val x = tytree' db t
       in (closecycles x x s; x) end (* Alle Knoten die von s gebunden waren hier her biegen *)
  | tytree' db (Lambda(s,t)) =
       let val x = tytree' db t
       in bindback x s; Ref(N(L lambda_,[x]))
       end
  | tytree' db (Id s) = (case dbgetidfromname s db of      (*Wenn ein Baum mit diesem Namen in der Datenbank ist*)
                            NONE => Ref(N(L(id_ s),nil))   (*dann diesen nehmen. *)
                          |(SOME id) => Ref(N(P id,nil)))
fun tytree db t = toDBLTree(tytree' db t)


(*****************************************************************

  findcycles: Liefert eine Menge der Knoten zurck, an denen
              ein Zyklus beginnt.

*****************************************************************)
fun findcycles' done s t = 
 if member(!done,t) then s:= add(!s,t)
 else (done:=add(!done,t); 
   case !t of 
     E => nil
   |(N(l,tr)) => (List.map (findcycles' done s) tr);
   done:=delete(!done,t))

fun findcycles t = let val c = Ref (Splayset.empty Ref.cmp)
                   in findcycles' (Ref (Splayset.empty Ref.cmp)) c t; !c end

(*****************************************************************

  tytree2ty: Wandelt einen zylischen Typenbaum in einen 
             Typ um.

*****************************************************************)
fun tytree2ty' db cycles bind i my l (t:TyLabel dbsig Tree Ref) = 
    let val b = Splaymap.peek(bind,t)        (* Nachsehen, ob der Baum schon  *)
                                             (* an einen Namen gebunden wurde *)
	val in_db = dbgetid (toDBLTree t) db (* Nachschauen, ob der Baum      *)
                                             (* schon in der Datenbank ist    *)
	val db_name = case in_db of NONE      => NONE
                                  | (SOME id) => dbgetname id db
	val in_cycle = member(cycles,t)      (* Beginnt hier ein Zyklus?      *)
    in
	case db_name of
	    (SOME s) => Id s                 (* wenn der Baum einen Namen hat,*)
                                             (* dann diesen zurckgeben       *)
	  | NONE =>
		if in_cycle andalso b=NONE then        (* Andernfalls nachsehen, ob wir ein *)
		    let val s = "t"^Int.toString(!i);  (* neues My erstellen mssen. *)
		    in  i:= !i+1;
			My(s,tytree2ty' db cycles (insert(bind,t,s)) i true l t)
		    end
		else if in_cycle andalso not my then Id (valOf b)  (* Baum ist durch ein My gebunden *)
		     else (case !t of                              (* Hier gehts in die Rekursion    *)
			       E => raise Error "Error in tytree2ty': unexpected E"
			     |(N(L null_,nil)) => Null
			     |(N(L unit_,nil)) => Unit
			     |(N(L add_,[t1,t2])) =>
				   Add(tytree2ty' db cycles bind i false l t1,
				       tytree2ty' db cycles bind i false l t2)
			     |(N(L arrow_,[t1,t2])) =>
				   Arrow(tytree2ty' db cycles bind i false l t1,
					 tytree2ty' db cycles bind i false l t2)
			     |(N(L mul_,t1::t2::tr)) =>
				   Mul(List.map (tytree2ty' db cycles bind i false l) (t1::t2::tr))
			     |(N(L(id_ s),nil)) =>
				   Id s
			     |(N(L(lambda_),[t1])) =>
				   Lambda("x"^Int.toString l,
					  tytree2ty' db cycles bind i false (l+1) t1)
			     |(N(L(back_ l'),nil)) =>
				   Id("x"^Int.toString(l-l'-1))
			     | _ => raise Error "tytree2ty': invalid type")
    end
fun tytree2ty db t = tytree2ty' db (findcycles t) (mkDict Ref.cmp) (Ref 0) false 0 t

fun tytree2tystring db t = ty2str (tytree2ty db t)

end
